import collections.abc
import sys

from django.contrib.admin.views.main import PAGE_VAR
from django.utils.functional import cached_property

from .cursor import Cursor, decode_cursor


class AdminCursorPaginator:
    def __init__(self, request, queryset, per_page, ordering_field):
        self.request = request
        self.queryset = queryset
        self.per_page = int(per_page)
        self.ordering_field = ordering_field
        self.ordering_field_name = ordering_field.lstrip('-')

    def page(self, _):
        return self.current_page

    @cached_property
    def current_page(self):
        encoded_cursor = self.request.GET.get(PAGE_VAR, '')
        cursor = decode_cursor(encoded_cursor)
        cur_number = cursor.number if cursor else 1
        is_reverse = bool(cursor and cursor.is_reverse)

        ordering_field = self.ordering_field
        if is_reverse:
            ordering_field = self._reverse_ordering(ordering_field)

        qs = self.queryset.order_by(ordering_field)
        if cursor:
            qs = self._apply_filter(qs, cursor)

        object_list = list(qs[: self.per_page + 1])
        has_more = len(object_list) > self.per_page
        if has_more:
            object_list = object_list[:-1]
        if is_reverse:
            object_list.reverse()

        prev_cursor = None
        if object_list and cur_number > 2:
            prev_cursor = Cursor(
                value=self._get_value_from_instance(object_list[0]),
                is_reverse=1,
                number=cur_number - 1,
            )

        next_cursor = None
        if object_list and (is_reverse or has_more):
            next_cursor = Cursor(
                value=self._get_value_from_instance(object_list[-1]),
                is_reverse=0,
                number=cur_number + 1,
            )

        return AdminCursorPage(object_list, self, cur_number, prev_cursor, next_cursor)

    @cached_property
    def count(self):
        return sys.maxsize

    def _apply_filter(self, qs, cursor):
        direction = 'gt' if cursor.is_reverse == self.ordering_field.startswith('-') else 'lt'
        lookup = '%s__%s' % (self.ordering_field_name, direction)
        return qs.filter(**{lookup: cursor.value})

    def _get_value_from_instance(self, instance):
        parts = self.ordering_field_name.split('__')
        attr = instance
        while parts:
            attr = getattr(attr, parts[0])
            parts.pop(0)
        return attr

    @classmethod
    def _reverse_ordering(cls, field):
        return field[1:] if field.startswith('-') else ('-' + field)


class AdminCursorPage(collections.abc.Sequence):
    def __init__(self, object_list, paginator, number, prev_cursor, next_cursor):
        self.object_list = object_list
        self.paginator = paginator
        self.number = number
        self.prev_cursor = prev_cursor
        self.next_cursor = next_cursor

    def __repr__(self):
        return '<Page starting from [%s]>' % (repr(self.object_list[0]) if self.object_list else None)

    def __len__(self):
        return len(self.object_list)

    def __getitem__(self, key):
        return self.object_list.__getitem__(key)

    def has_next(self):
        return self.next_cursor is not None

    def has_previous(self):
        return self.number > 1

    def has_other_pages(self):
        return self.has_previous() or self.has_next()



from base64 import b64decode, b64encode
from collections import namedtuple
from datetime import datetime


Cursor = namedtuple('Cursor', ['value', 'is_reverse', 'number'])

_delimiter = '|||'


def encode_cursor(cursor):
    parts = [encode_value(x) for x in cursor]
    return b64encode(_delimiter.join(parts).encode('utf8')).decode('ascii')


def decode_cursor(encoded_cursor):
    try:
        cursor_str = b64decode(encoded_cursor.encode('ascii')).decode('utf8')
        parts = cursor_str.split(_delimiter)
        return Cursor(
            value=decode_value(parts[0]),
            is_reverse=bool(parts[1]),
            number=int(parts[2]),
        )
    except (TypeError, ValueError):
        return None


def encode_value(val):
    if not val:
        return ''
    # datetime.__str__ applies `isoformat()` under the hood,
    # but we do this explicitly for the sake of clarity
    if isinstance(val, datetime):
        return val.isoformat()
    return str(val)


def decode_value(val):
    if not val:
        raise ValueError
    try:
        return _decode_datetime(val)
    except ValueError:
        return val


def _decode_datetime(val):
    try:
        return datetime.fromisoformat(val)
    except AttributeError:
        pass

    # back compat for python versions prior to 3.7
    # (they don't have `fromisoformat` method)
    from django.utils.dateparse import parse_datetime

    res = parse_datetime(val)
    if res:
        return res

    raise ValueError



import django

from django.contrib import admin

from .paginator import AdminCursorPaginator
from .views import CursorPaginatorChangeList


class CursorPaginatorAdmin(admin.ModelAdmin):
    # django built-in params
    change_list_template = 'admin_cursor_paginator/change_list.html'
    sortable_by = ()
    show_full_result_count = False
    actions_selection_counter = False

    # specific cursor paginator params
    cursor_ordering_field = '-pk'
    show_query_result_count = True

    def get_paginator(self, request, queryset, per_page, orphans=0, allow_empty_first_page=True):
        # django built-in `history_view` shares `ModelAdmin` paginator,
        # so we need to pass original one for this specific case
        if django.VERSION >= (4, 1):
            from django.contrib.admin.models import LogEntry
            if queryset.model is LogEntry:
                return self.paginator(queryset, per_page, orphans, allow_empty_first_page)

        return AdminCursorPaginator(request, queryset, per_page, self.cursor_ordering_field)

    def get_changelist(self, request, **kwargs):
        return CursorPaginatorChangeList



from django.contrib.admin.views.main import ChangeList


class CursorPaginatorChangeList(ChangeList):
    def get_ordering(self, *args, **kwargs):
        return [self.model_admin.cursor_ordering_field]

    def get_results(self, *args, **kwargs):
        super().get_results(*args, **kwargs)
        if self.query and self.model_admin.show_query_result_count:
            self.result_count = self.queryset.count()
        else:
            # small hack to prevent full count (which is unknown) from being shown in `search_form.html`
            self.full_result_count = self.result_count



#!/usr/bin/env python3

import os

from setuptools import setup


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):  # __version__ = '0.1.2'
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError('Unable to find version string.')


setup(
    version=get_version('admin_cursor_paginator/version.py'),
)



from django.contrib.admin.views.main import PAGE_VAR
from django.template.defaulttags import register

from ..cursor import encode_cursor


@register.inclusion_tag('admin_cursor_paginator/pagination.html', takes_context=False)
def admin_cursor_pagination(cl):
    page = cl.paginator.current_page
    cur_number = page.number

    first_page_url = ''
    prev_page_url = ''
    next_page_url = ''

    if cur_number > 1:
        first_page_url = cl.get_query_string(remove=[PAGE_VAR])

    if cur_number == 2:
        prev_page_url = first_page_url
    elif page.prev_cursor:
        prev_cursor_repr = encode_cursor(page.prev_cursor)
        prev_page_url = cl.get_query_string(new_params={PAGE_VAR: prev_cursor_repr})

    if page.next_cursor:
        next_cursor_repr = encode_cursor(page.next_cursor)
        next_page_url = cl.get_query_string(new_params={PAGE_VAR: next_cursor_repr})

    return {
        'cl': cl,
        'cur_number': cur_number,
        'first_page_url': first_page_url,
        'prev_page_url': prev_page_url,
        'next_page_url': next_page_url,
    }
